/****************************************************************************
 *
 * Julia set test program for PS3 cell architecture.
 *
 * (c) 2007 Ranulf Doswell <dev@ranulf.net>
 *
 * I've only been programming the PS3 for just over 24 hours and here is my
 * first attempt at just the SPUs, DMA transfers and hand-optimising SBU
 * assembler (such a nice processor to experiment with)...
 *
 * Please feel free to use this as a basis for your experimentations, but if
 * you come up with anything cool, please let me know!
 *
 ****************************************************************************/

#include "fb.h"
#include "general.h"

FrameBuffer::FrameBuffer() {
    struct ps3fb_ioctl_res res;

	consoleFD = open("/dev/fb0", O_RDWR);
    if (consoleFD<0)
        throw InitialisationError();

    if (ioctl(consoleFD, PS3FB_IOCTL_SCREENINFO, (unsigned long)&res)==-1)
        throw InitialisationError();

    if (res.num_frames<2)
        throw InitialisationError();

    width = res.xres;
    height = res.yres;

    left = res.xoff;
    top = res.yoff;

    right = width-left;
    bottom = height-top;

    length = width * height * 4 * res.num_frames;
    addr = mmap(0, length, PROT_WRITE, MAP_SHARED, consoleFD, 0);
    addr2 = (void*) ( ((char*)addr) + width*height*4 );
    frame = false;

    printf("Screen resolution is %dx%d, viewport (%d,%d)-(%d,%d)\n",
        width, height, left, top, right, bottom);
    printf("Buffers are at %08x and %08x, total length is %d\n", 
        addr, addr2, length);

    ioctl(consoleFD, PS3FB_IOCTL_ON, 0);
    memset(addr, 0, length);
  
    flip();

    gettimeofday(&tv, NULL);
    framecount = 0;

	switchMode(true);
}

FrameBuffer::~FrameBuffer() {
    ioctl(consoleFD, PS3FB_IOCTL_OFF, 0);
    munmap(addr, length);
    close(consoleFD);
	switchMode(false);
}

void FrameBuffer::switchMode(bool gfxMode) {
	int fd = open("/dev/console", O_NONBLOCK);
        if (fd >= 0) {
		ioctl(fd, KDSETMODE, gfxMode ? KD_GRAPHICS : KD_TEXT);
                close(fd);
	}
}

void* FrameBuffer::getFrameBuffer() {
    return current;
}

void* FrameBuffer::getOnScreenFrameBuffer() {
    return frame ? addr2 : addr;
}

void FrameBuffer::flip() {
    uint32_t framesel = frame ? 1 : 0;
    ioctl(consoleFD, PS3FB_IOCTL_FSEL, (unsigned long)&framesel);
    frame = !frame;
    current = frame ? addr : addr2;
    framecount++;
}

void FrameBuffer::vsync() {
    uint32_t crt = 0;
    ioctl(consoleFD, FBIO_WAITFORVSYNC, (unsigned long)&crt);
} 

float FrameBuffer::getFPS() {
    struct timeval tv2;
    gettimeofday(&tv2, NULL);

    int seconds = tv2.tv_sec - tv.tv_sec;
    int usec = tv2.tv_usec - tv.tv_usec;
    seconds += (usec<0) ? -1 : 0;
    usec += (usec<0) ? 1000000 : 0;

	float normalised = ((float)seconds) + ((float)usec)/1000000;
    return framecount/normalised;
}

int FrameBuffer::getFramecount() {
    return framecount;
}
