/****************************************************************************
 *
 * Julia set test program for PS3 cell architecture.
 *
 * (c) 2007 Ranulf Doswell <dev@ranulf.net>
 *
 * I've only been programming the PS3 for just over 24 hours and here is my
 * first attempt at just the SPUs, DMA transfers and hand-optimising SBU
 * assembler (such a nice processor to experiment with)...
 *
 * Please feel free to use this as a basis for your experimentations, but if
 * you come up with anything cool, please let me know!
 *
 ****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <linux/kd.h>
#include <linux/joystick.h>

#include "general.h"

#include "joystick.h"

Joystick::Joystick() {
	joyFD = open("/dev/js0", O_RDONLY);
	if (joyFD<0)
		throw InitialisationError();
	fcntl(joyFD, F_SETFL, O_NONBLOCK);

	for (int i=0; i<4; i++)
		axis[i] = 0;

	buttons = 0;
}

void Joystick::update() {
	struct js_event js;
	int r=0;
	while ((r=read(joyFD, &js, sizeof(struct js_event))) >0) {
		if (js.type & JS_EVENT_BUTTON) {
			if (js.value)
				buttons |= (1<<js.number);
			else
				buttons &=~ (1<<js.number);
		}
		if (js.type & JS_EVENT_AXIS && js.number<4) {
			axis[js.number] = js.value;
		}
	}
}

int Joystick::getAllButtons() {
	update();
	return buttons;
}

int Joystick::getButton(int b) {
	update();
	return (buttons>>b) & 1;
}

int Joystick::getAxis(int a) {
	update();
	return axis[a];
}

Joystick::~Joystick() {
	close(joyFD);
}
