/****************************************************************************
 *
 * Julia set test program for PS3 cell architecture.
 *
 * (c) 2007 Ranulf Doswell <dev@ranulf.net>
 *
 * I've only been programming the PS3 for just over 24 hours and here is my
 * first attempt at just the SPUs, DMA transfers and hand-optimising SBU
 * assembler (such a nice processor to experiment with)...
 *
 * Please feel free to use this as a basis for your experimentations, but if
 * you come up with anything cool, please let me know!
 *
 ****************************************************************************/

//Pull in DMA commands
#include <spu_mfcio.h>
#include "protocol.h"

typedef unsigned int u32;
typedef u32* EffectiveAddress;

inline int julia(float x_,float y_, float x0_, float y0_,
								float stepx_xinc, float stepx_yinc) {
	register int it=0;
	register float x = x_;
	register float y = y_;
	register float x0 = x0_;
	register float y0 = y0_;

	while (it++<255) {
		register float x2 = x*x;
		register float y2 = y*y;
		y = 2*x*y + y0;
		x = x2-y2 + x0;
		if (x2+y2 > 4.0)
			return it;
	}

	return 0;
}

void juliaLine(u32* line, int width, float fx, float fy, float x0, float y0, float stepx_xinc, float stepx_yinc) {
	int x;
	for (x=0; x<width; x++) {
		int i = julia(fx,fy,x0,y0,stepx_xinc,stepx_yinc);
		fx += stepx_xinc;
		fy += stepx_yinc;
	
		int r = i<<4;
		int g = i<<2;
		int b = i;
		line[x] = ((r&255)<<16) | ((g&255)<<8) | (b&255);
	}
}

int main(unsigned long long spe_id, unsigned long long program_data_ea, unsigned 
long long env) {
	program_data pd __attribute__((aligned(16)));

	int tag_id = 0;

#ifdef MESSAGES
	for (;;) {
		int msg = spu_read_in_mbox();
		if (msg<0) {
			printf("[%lld] terminating\n", spe_id);
			break;
		}

		tag_id= (tag_id+1)%32;
#endif

		//READ DATA IN
		mfc_get(&pd, program_data_ea, sizeof(pd), tag_id, 0, 0);
		mfc_write_tag_mask(1<<tag_id);
		int mask = mfc_read_tag_status_any();
	
		//PROCESS DATA
	
#ifdef MESSAGES
		/*
		printf("%f,%f %f,%f %f,%f %f,%f\n%08lx %dx%d %d,%d,%d\n",
			pd.corner_x, pd.corner_y, pd.x0, pd.y0, 
			pd.stepx_xinc, pd.stepx_yinc, pd.stepy_xinc, pd.stepy_yinc,
			pd.framebuffer, pd.width, pd.height, pd.start, pd.count, pd.tag);
		*/

		spu_write_out_mbox(42);
#endif
	
		int length = pd.width*pd.count*4;
		u32* buffer = malloc(length);
		EffectiveAddress frame = pd.framebuffer;
		int linelength = pd.width * sizeof(u32);
	
		float f_x = pd.corner_x + pd.start * pd.stepy_xinc;
		float f_y = pd.corner_y + pd.start * pd.stepy_yinc;
	
		int x,y;
		for (y=0; y<pd.count; y++) {
			float fx = f_x;
			float fy = f_y;
			f_x += pd.stepy_xinc;
			f_y += pd.stepy_yinc;
	
			u32* line = buffer + y*pd.width;
			EffectiveAddress dest = frame + (y+pd.start)*pd.width;
	
			juliaLine(line, pd.width, fx, fy, pd.x0, pd.y0, pd.stepx_xinc, pd.stepx_yinc);
	
			tag_id= (tag_id+1)%32;

			unsigned long long ea = (unsigned long long) ((unsigned long)dest);
			mfc_put(line, ea, linelength, tag_id, 0, 0);
			mfc_write_tag_mask(1<<tag_id);
			int mask = mfc_read_tag_status_any();
		}
	
		free(buffer);
#ifdef MESSAGES
	}
#endif
	
	return 0;
}
