/****************************************************************************
 *
 * Julia set test program for PS3 cell architecture.
 *
 * (c) 2007 Ranulf Doswell <dev@ranulf.net>
 *
 * I've only been programming the PS3 for just over 24 hours and here is my
 * first attempt at just the SPUs, DMA transfers and hand-optimising SBU
 * assembler (such a nice processor to experiment with)...
 *
 * Please feel free to use this as a basis for your experimentations, but if
 * you come up with anything cool, please let me know!
 *
 ****************************************************************************/

#include "fb.h"
#include "spu.h"
#include "protocol.h"
#include "joystick.h"

#define DEFINE_ACTIVE_PROGRAM(x) extern spe_program_handle_t x; \
							     spe_program_handle_t* program = &x;

#ifndef UNOPT
DEFINE_ACTIVE_PROGRAM(optimised_handle)
#else
DEFINE_ACTIVE_PROGRAM(julia_spe_handle)
#endif

#define NUM_SPUS 6
#define FRAME_STEP 150
#define FRAME_START 8010
#define FRAME_END 182500
#define VSYNC

inline void* aligned_malloc(int length) {
	void* p = malloc(length+15);
	unsigned long ptr = (unsigned long) p;
	void* p2 = (void*) ( (ptr+15) & ~15);
	//printf("Aligning %08lx -> %08lx\n", p, p2);
	//char** pp =(char**)p2;
	//p2[-1]=p;
	return p2;
}

inline void aligned_free(void *p) {
//	char** pp =(char**)p;
//	free(p[-1]);
}

int main(int argc, char *argv[]) {
    uint32_t frame = 0;
    struct timeval tv;
    uint32_t time;

	if (sizeof(program_data) & 15) {
		printf("FATAL error: program_data structure size is %d bytes long.\n",
			sizeof(program_data));
		exit(1);
	}

	Joystick joy;

    FrameBuffer fb;

	Spu* spu[NUM_SPUS];
	program_data* data[NUM_SPUS];
		
	for (int s=0; s<NUM_SPUS; s++) {
		data[s]=(program_data*)aligned_malloc(sizeof(program_data));
		spu[s]=new Spu(program, data[s], NULL);
	}

	struct timeval tv1;
	struct timeval tv2;
	gettimeofday(&tv1, NULL);
	gettimeofday(&tv2, NULL);
	int wtf=0;

	float target_x = 0.32;
	float target_y = 0.40;

	float x0 = 0.285;
	float y0 = 0.01;

	float xinc = 0.001;
	float yinc = 0.001;

	while (!joy.getButton(JOYSTICK_CIRCLE)) { //for (int s=FRAME_START; s<FRAME_END; s+=FRAME_STEP) {
		gettimeofday( (wtf++&1) ? &tv1 : &tv2, NULL);
		
		int seconds = tv2.tv_sec - tv1.tv_sec;
		int usec = tv2.tv_usec - tv1.tv_usec;
		seconds += (usec<0) ? -1 : 0;
		usec += (usec<0) ? 1000000 : 0;
		
		float normalised = fabs(((float)seconds) + ((float)usec)/1000000);

/*
		printf("Frame %d took %2.8f. Joystick buttons: %08x %d,%d %d,%d\n", 
			s, normalised, joy.getAllButtons(),
			joy.getAxis(0), joy.getAxis(1), joy.getAxis(2), joy.getAxis(3) );
*/

		target_x += normalised*xinc*.01*joy.getAxis(2);
		target_y += normalised*yinc*.01*joy.getAxis(3);
		
		x0 += normalised*.00000003*joy.getAxis(0);
		y0 += normalised*.00000003*joy.getAxis(1);

		if (joy.getButton(JOYSTICK_R1)) {
			xinc *= (1.0+0.8*normalised);
		}
		if (joy.getButton(JOYSTICK_R2)) {
			xinc /= (1.0+0.8*normalised);
		}
		yinc = xinc;
		
//		printf("Now at %2.8f,%2.8f\n", target_x, target_y);
	
        fb.flip();

		int block = 15;

		int sindex = 0;
	
		for (int y=0; y<fb.height;y+=block) {
			data[sindex]->x0 = x0;
			data[sindex]->y0 = y0;

//			data[sindex]->x0 = -1.0 / 1.6 * (s+1)/4500.0;
//			data[sindex]->y0 = -1.0 / 1.4 * (s+1)/3500.0;
//			data[sindex]->y0 = -1.0 + (s+1)/19700.0;

			data[sindex]->stepx_xinc = xinc; //(5.0/fb.width) / (s+1);
			data[sindex]->stepy_yinc = yinc; //data[sindex]->stepx_xinc;
			data[sindex]->stepx_yinc = 0.0;
			data[sindex]->stepy_xinc = 0.0;
			data[sindex]->corner_x = target_x - (fb.width/2)*data[sindex]->stepx_xinc;
			data[sindex]->corner_y = target_y - (fb.height/2)*data[sindex]->stepy_yinc;
	/*
			printf("s,x0,x0 = %5d,%3.10f,%3.10f [%0.5f]\n",s,
				data[sindex]->x0, data[sindex]->y0, normalised);
	*/
			data[sindex]->framebuffer = fb.getFrameBuffer();
			data[sindex]->width = fb.width;
			data[sindex]->height = fb.height;

			data[sindex]->start = y;
			data[sindex]->count = (y+block>fb.height) ? (fb.height-y) : block;
			data[sindex]->tag = 0;
			spu[sindex]->write(sindex);
			spu[sindex]->read();
			sindex = (sindex+1) % NUM_SPUS;
		}
		for (int pause=0; pause<1; pause++) {
#ifdef VSYNC
           fb.vsync();
#endif
}
    }

	for (int q=0; q<NUM_SPUS; q++) {
		if (spu[q]) {
			spu[q]->write(-1);
			spu[q]->join();
			delete spu[q];
		}
		//aligned_free(data[q]);
	}

    printf("Average framecount over %d frames: %.1f FPS\n", fb.getFramecount(),
        fb.getFPS());
}
